/* 
 * FILE --- mlp.hh
 *
 * PURPOSE --- Header file for a class that implements a Multilayer 
 * Perceptron (MLP) neural network (NN). 
 * The user specifies the number of inputs, the number of outouts and the 
 * number of the neurons in the hidden layer of the perceptron. It is possible
 * to choose different activation functions for the neurans. The user can 
 * either specify the input and output weigth matrices and biaces, or let them 
 * be choosen randomly within the [-1.0, 1.0] range.
 *
 * This utility uses the matrix tool for matrix maltiplications.
 *
 */


// Prevent Multiple Inclusion
#ifndef MLP_HH
#define MLP_HH


// Include Files
#include <stdlib.h>
#include "matrix.hh"


// Local Definitions and Macros
#define MLP_NO_ERROR 0
#define MLP_MEM_ERROR 1

#define LOGSIG 0  // Log-Sigmoid activation function
#define TANSIG 1  // Hyperbolic Tangent Sigmoid activation function
#define HARDLIM 2  // Hard limiting function (like a step)
#define HARDLIMS 3  // Hard limiting function (like a step) symmetric wrt origin
#define PURELIN 4  // Purely linear function 
#define SATLIN 5  // Linear function with saturation 
#define SATLINS 6  // Linear function with saturation symmetric wrt origin
#define POSLIN 7  // Positive linear function 


class MLP
{
public:
  
  // Constructor 
  MLP(int _num_of_inputs, int _num_of_outputs, int _num_of_neurons);

  // Destructor
  ~MLP();

  // Functions to set the input weights and biases
  void set_input_weights(double _low, double _high);
  void set_input_weights(const double _in_weights);
  void set_input_weights(const double *_in_weights);
  void set_input_weights(const double **_in_weights);
  void set_input_biases(double _low, double _high);
  void set_input_biases(const double _in_biases);
  void set_input_biases(const double *_in_biases);

  // Functions to set the output weights and biases
  void set_output_weights(double _low, double _high);
  void set_output_weights(const double _out_weights);
  void set_output_weights(const double *_out_weights);
  void set_output_weights(const double **_out_weights);
  void set_output_biases(double _low, double _high);
  void set_output_biases(const double _out_biases);
  void set_output_biases(const double *_out_biases);

  // Set or return the type of the activation functions
  void set_activation_function_type(int _actv_func_type);
  int get_activation_function_type(void) { return actv_func_type; }

  // Functions to set and return the adaptation gain
  void set_eta(double _eta) { eta = _eta; };
  double get_eta(void) { return eta; };

  // Function to return the norm-square of the regressor
  double get_phi_norm(void) { return phi_norm; };

  // The function that given the inputs computes the the outputs 
  // of the MLP. 
  // input = an array of inputs of size _num_of_inputs
  // output = an array of outputs of size _num_of_outputs
  void control(const double *input, double *output);

  // The function that given the error updates the output parameters 
  void update_output_params(const double *error);
  void update_output_params(const double *error, double norm2);
  void update_output_params(const double *error, double norm2, double uu);
  void bound_output_weights(double _low, double _high);
  void bound_output_biases(double _low, double _high);

  int mlp_err; // Variable to hold the error condition

private:
  // Allocate/free memory for the arrays
  void allocate_memory(void);
  void free_memory(void);

  // Function that calculates the activation of each neuron
  void get_activation(void);
  // Function that transfers the outputs from a matrix to the output variable
  void get_output(double *_y);

  // Pointer to the currently chosen activation function for the neurons
  double (*actv_func) (const double input);

  int num_of_inputs;  // # of inputs (to be specified by the user)
  int num_of_outputs;  // # of outputs (to be specified by the user)
  int num_of_neurons;  // # of neurons (to be specified by the user)

  int actv_func_type;  // Type of activation function
  double phi_norm; // Variable to hold the norm of phi 
  double eta; // Adaptation gain

  // The weigth and bias matrices
  matrix *Wi;  // Input weigth matrix
  matrix *bi;  // Input bias matrix
  matrix *Wo;  // Output weigth matrix
  matrix *bo;  // Output bias matrix

  // Convaniance variables
  matrix *u; // Input vector
  matrix *x; // Intermediate vector
  matrix *phi; // Intermediate vector
  matrix *y; // Output vector
  matrix *e; // Error vector 
};

// Activation Functions for the neurons
inline double logsig(const double _input);
inline double tansig(const double _input);
inline double hardlim(const double _input);
inline double hardlims(const double _input);
inline double purelin(const double _input);
inline double satlin(const double _input);
inline double satlins(const double _input);
inline double poslin(const double _input);

#endif // MLP_HH
