/* 
 * FILE --- mlp_ex1.cc
 *
 * PURPOSE --- To provide an example use of the class for 4'th
 * order Runge-Kutta simulation, RK4_SIM, and the class for implementing
 * n-input m-output Multilayer Perceptron (MLP) neural network (NN).
 *
 * Implements the main function of a class for performing a simulation of
 * a simple cargo ship whose steering is controlled by a MLP controller.
 *
 */

// Include Files
#include <stdlib.h>
#include <stdio.h>

#include "mlp.hh"
#include "tanker_sim.hh"

// Local Definitions and Macros
#define PI 3.14159265359

void main(void) 
{
  double Ts = 10.0;             // Sampling time
  double t0 = 0.0;              // Initial time 
  double tmax = 4000.0;         // Total Simulation period

  double y[3] = {0.0, 0.0, 0.0};  // Initial conditions    

  // Weights and biases for the MLP
  double bi[2] = {-200.0*PI/180.0, +200.0*PI/180.0};  // Input biases
  double Wi[2] = {10.0, 10.0};  // Input weigths
  double bo[1] = {80.0*PI/180.0};  // Output biases
  double Wo[2] = {-80.0*PI/180.0, -80.0*PI/180.0}; // Output weigths

  // # of simulation steps to be performed by the RK4 algorith
  // in one sampling interval
  int sim_per_cycle = 10;  

  // Create the tanker object
  TANKER_SIM my_tanker(Ts,sim_per_cycle,t0,(double *)y);

  // Create the MLP object
  MLP my_mlp_ctrl(1, 1, 2);

  double ctrl = 0.0;             // Control variable
  double error = 0.0;            // Error variable
  double ref = 0.0;  // Reference input variable

  FILE *fp;  // To save data

  if (!(fp = fopen("tanker_mlp_out.dat","w"))) {
    printf("Can't open file.\n");
    exit(1);
  }

  // Set the MLP NN controller weigths and biases
  my_mlp_ctrl.set_input_weights(Wi);
  my_mlp_ctrl.set_input_biases(bi);
  my_mlp_ctrl.set_output_weights(Wo);
  my_mlp_ctrl.set_output_biases(bo);

  // Set the activation functions to be Log-Sigmaids
  my_mlp_ctrl.set_activation_function_type(LOGSIG);

  /* Simulate for tmax seconds. */
  while(my_tanker.current_time < tmax) 
    {
      // Update the reference input
      if (my_tanker.current_time < 100.0)
  	ref = 0;   // Request heading of 0 deg
      else if (my_tanker.current_time < 2000.0)
 	ref = 45.0*(PI/180.0); // Request heading of 45 deg
      else
  	ref = 0;   // Request heading of 0 deg
      
      // Calculate the control input
      error = (ref -  my_tanker.plant_state[0]);
      my_mlp_ctrl.control(&error, &ctrl);
      
      // Simulate the tanker ship dynamics
      my_tanker.simulate(ctrl);
      
      // Save the data
      fprintf(fp,"%f %f %f %f %f %f \n", my_tanker.current_time, ref,
  	      my_tanker.plant_state[0], my_tanker.plant_state[1],
  	      my_tanker.plant_state[2], my_tanker.ctrl_input[0]);
    }
  
  
  // The code here can be used to view the function that the MLP implements
  //    error = -80.0*PI/180.0;
  //    for(int ii=0; ii < 160; ii++)
  //    {
  //      my_mlp_ctrl.control(&error, &ctrl);
  //      fprintf(fp,"%f %f \n", error, ctrl);
  //      error += (2.0*80.0*PI)/(180.0*160.0);
  //    }
    
  fclose(fp);

  printf("Simulation completed ...\n");
}
