/* 
 * FILE --- mlp_ex3.cc
 *
 * PURPOSE --- To provide an example use of the class for 4'th
 * order Runge-Kutta simulation, RK4_SIM, and the class for implementing
 * n-input m-output Multilayer Perceptron (MLP) neural network (NN).
 *
 * Implements the main function of a class for performing a simulation of a simple 
 * surge tank level regulation controlled by a indirect adaptive MLP controller.
 *
 * NOTE: If you are using the RBF NN as an indirect adaptive controller, then the 
 * control input to your controlled system should be one dimensional (for this tool).
 *
 */

// Include Files
#include <stdlib.h>
#include <stdio.h>

#include "mlp.hh"
#include "tank_sim.hh"

// Local Definitions and Macros
#define PI 3.14159265359
#define NUM_NEURONS 10

void main(void) 
{
  double Ts = 0.1;             // Sampling time
  double t0 = 0.0;              // Initial time 
  double tmax = 1000.0;         // Total Simulation period

  double y[1] = {1.0};  // Initial conditions    

  // Weights and biases for the MLP
  double bo = 0.0;  // Output biases
  double Wo = 0.0;  // Output weigths

  // # of simulation steps to be performed by the RK4 algorith
  // in one sampling interval
  int sim_per_cycle = 10;  

  // Create the tank object
  TANK_SIM my_tank(Ts,sim_per_cycle,t0,(double *)y);

  // Create the MLP object
  MLP alpha_hat(1, 1, NUM_NEURONS);
  MLP beta_hat(1, 1, NUM_NEURONS);

  double ctrl = 0.0;             // Control variable
  double input_to_mlp = 0.0;  // Error variable
  double error = 0.0;            // Error variable
  double epsilon = 0.001;  // Deadband variable
  double ref = 0.0;  // Reference input variable
  double eta = - 1.25; // Adaptation gain
  double a_hat = 0.0, b_hat = 0.0;
  double beta0 = 0.25, beta1 = 0.5;

  FILE *fp;  // To save data

  if (!(fp = fopen("tank_mlp_ind_out.dat","w"))) {
    printf("Can't open file.\n");
    exit(1);
  }

  // Set the MLP NN controller weigths and biases
  alpha_hat.set_input_weights(-10.0, 10.0);
  alpha_hat.set_input_biases(-200.0*PI/180.0, +200.0*PI/180.0);
  alpha_hat.set_output_weights(Wo);
  alpha_hat.set_output_biases(bo);
  beta_hat.set_input_weights(-10.0, 10.0);
  beta_hat.set_input_biases(-200.0*PI/180.0, +200.0*PI/180.0);
  beta_hat.set_output_weights(beta0+0.5*(beta1-beta0)); // all output parameters b/n b0 anb b1
  beta_hat.set_output_biases(beta0+0.5*(beta1-beta0));

  // Set the activation functions to be Log-Sigmaids
  alpha_hat.set_activation_function_type(LOGSIG);
  beta_hat.set_activation_function_type(LOGSIG);

  // Set the adaptation gain 
  alpha_hat.set_eta(eta);
  beta_hat.set_eta(eta);

  /* Simulate for tmax seconds. */
  while(my_tank.current_time < tmax) 
    {
      // Update the reference input
      if (my_tank.current_time < 100.0)
  	ref = 3.25;   // Request level 3.25
      else if (my_tank.current_time < 300.0)
 	ref = 0.25; // Request level of 0.25 
      else if (my_tank.current_time < 500.0)
 	ref = 3.25; // Request level of 3.25 
      else if (my_tank.current_time < 700.0)
 	ref = 0.25; // Request level of 0.25 
      else if (my_tank.current_time < 900.0)
 	ref = 3.25; // Request level of 3.25 
      else
  	ref = 3.0;   // Request level of 3.0
      
      // Calculate the input to the controller
      input_to_mlp = my_tank.plant_state[0];
      
      // Calculate the control input
      alpha_hat.control(&input_to_mlp, &a_hat);
      beta_hat.control(&input_to_mlp, &b_hat);
      ctrl = (- a_hat + ref)/b_hat;  

      // Simulate the tank ship dynamics
      my_tank.simulate(ctrl);

      // Calculate the error
      error = ref - my_tank.plant_state[0];

      // Pass the error through a deadband
      if (error > epsilon)
	{
	  error -= epsilon;
	}
      else if (error < - epsilon)
	{
	  error += epsilon;
	}

      // Update the output parameters of the controller
      // NOTE: The ataptation in the indirect case is different. Since we use two approximators
      // insted of one, we need to pass information from one approximator to the other in
      // order for the parameter update to guarantee stability. Moreover, we need to pass
      // the control input to the approximator of "beta" (the function multiplying the control
      // input, since it (the control) is part of its regression vector.
      alpha_hat.update_output_params(&error, beta_hat.get_phi_norm());
      beta_hat.update_output_params(&error, alpha_hat.get_phi_norm(), ctrl);
      beta_hat.bound_output_weights(beta0, beta1);
      beta_hat.bound_output_biases(beta0, beta1);

      // Save the data
      fprintf(fp,"%f %f %f %f %f %f \n", my_tank.current_time, ref,
      	      my_tank.plant_state[0], my_tank.ctrl_input[0]);
    }
  
  fclose(fp);

  printf("Simulation completed ...\n");
}
